@@ -80,9 +80,10 @@
 #![allow(clippy::float_cmp, clippy::manual_range_contains)]
 
 mod flamegraph;
+mod maybe_mut_ref;
 mod stats;
 
-pub use {egui, puffin};
+pub use {egui, maybe_mut_ref::MaybeMutRef, puffin};
 
 use egui::*;
 use puffin::*;
@@ -131,6 +132,7 @@ pub fn profiler_ui(ui: &mut egui::Ui) {
 pub struct GlobalProfilerUi {
     #[cfg_attr(feature = "serde", serde(skip))]
     global_frame_view: GlobalFrameView,
+
     pub profiler_ui: ProfilerUi,
 }
 
@@ -142,7 +144,8 @@ impl GlobalProfilerUi {
     /// Returns `false` if the user closed the profile window.
     pub fn window(&mut self, ctx: &egui::CtxRef) -> bool {
         let mut frame_view = self.global_frame_view.lock();
-        self.profiler_ui.window(ctx, &mut frame_view)
+        self.profiler_ui
+            .window(ctx, &mut MaybeMutRef::MutRef(&mut frame_view))
     }
 
     /// Show the profiler.
@@ -150,7 +153,8 @@ impl GlobalProfilerUi {
     /// Call this from within an [`egui::Window`], or use [`Self::window`] instead.
     pub fn ui(&mut self, ui: &mut egui::Ui) {
         let mut frame_view = self.global_frame_view.lock();
-        self.profiler_ui.ui(ui, &mut frame_view);
+        self.profiler_ui
+            .ui(ui, &mut MaybeMutRef::MutRef(&mut frame_view));
     }
 
     /// The frames we are looking at.
@@ -175,6 +179,14 @@ impl AvailableFrames {
             slowest: frame_view.slowest_frames_chronological(),
         }
     }
+
+    fn all_uniq(&self) -> Vec<Arc<FrameData>> {
+        let mut all = self.slowest.clone();
+        all.extend(self.recent.iter().cloned());
+        all.sort_by_key(|frame| frame.frame_index());
+        all.dedup_by_key(|frame| frame.frame_index());
+        all
+    }
 }
 
 /// Multiple streams for one thread.
@@ -186,7 +198,7 @@ pub struct Streams {
 }
 
 impl Streams {
-    fn new(frames: &[Arc<FrameData>], thread_info: &ThreadInfo) -> Self {
+    fn new(frames: &[Arc<UnpackedFrameData>], thread_info: &ThreadInfo) -> Self {
         crate::profile_function!();
 
         let mut streams = vec![];
@@ -196,7 +208,10 @@ impl Streams {
             }
         }
 
-        let merges = puffin::merge_scopes_for_thread(frames, thread_info).unwrap();
+        let merges = {
+            puffin::profile_scope!("merge_scopes_for_thread");
+            puffin::merge_scopes_for_thread(frames, thread_info).unwrap()
+        };
         let merges = merges.into_iter().map(|ms| ms.into_owned()).collect();
 
         let mut max_depth = 0;
@@ -217,22 +232,22 @@ impl Streams {
 #[derive(Clone)]
 pub struct SelectedFrames {
     /// ordered, but not necessarily in sequence
-    pub frames: vec1::Vec1<Arc<FrameData>>,
+    pub frames: vec1::Vec1<Arc<UnpackedFrameData>>,
     pub raw_range_ns: (NanoSecond, NanoSecond),
     pub merged_range_ns: (NanoSecond, NanoSecond),
     pub threads: BTreeMap<ThreadInfo, Streams>,
 }
 
 impl SelectedFrames {
-    fn try_from_vec(frames: Vec<Arc<FrameData>>) -> Option<Self> {
+    fn try_from_vec(frames: Vec<Arc<UnpackedFrameData>>) -> Option<Self> {
         let frames = vec1::Vec1::try_from_vec(frames).ok()?;
         Some(Self::from_vec1(frames))
     }
 
-    fn from_vec1(mut frames: vec1::Vec1<Arc<FrameData>>) -> Self {
+    fn from_vec1(mut frames: vec1::Vec1<Arc<UnpackedFrameData>>) -> Self {
         puffin::profile_function!();
-        frames.sort_by_key(|f| f.frame_index);
-        frames.dedup_by_key(|f| f.frame_index);
+        frames.sort_by_key(|f| f.frame_index());
+        frames.dedup_by_key(|f| f.frame_index());
 
         let mut threads: BTreeSet<ThreadInfo> = BTreeSet::new();
         for frame in &frames {
@@ -257,7 +272,7 @@ impl SelectedFrames {
             }
         }
 
-        let raw_range_ns = (frames.first().range_ns.0, frames.last().range_ns.1);
+        let raw_range_ns = (frames.first().range_ns().0, frames.last().range_ns().1);
 
         Self {
             frames,
@@ -268,7 +283,7 @@ impl SelectedFrames {
     }
 
     pub fn contains(&self, frame_index: u64) -> bool {
-        self.frames.iter().any(|f| f.frame_index == frame_index)
+        self.frames.iter().any(|f| f.frame_index() == frame_index)
     }
 }
 
@@ -307,6 +322,10 @@ pub struct ProfilerUi {
 
     /// Used to normalize frame height in frame view
     slowest_frame: f32,
+
+    /// When did we last run a pass to pack all the frames?
+    #[cfg_attr(feature = "serde", serde(skip))]
+    last_pack_pass: Option<std::time::Instant>,
 }
 
 impl Default for ProfilerUi {
@@ -316,6 +335,7 @@ impl Default for ProfilerUi {
             view: Default::default(),
             paused: None,
             slowest_frame: 0.16,
+            last_pack_pass: None,
         }
     }
 }
@@ -330,7 +350,11 @@ impl ProfilerUi {
     /// If you want to control the window yourself, use [`Self::ui`] instead.
     ///
     /// Returns `false` if the user closed the profile window.
-    pub fn window(&mut self, ctx: &egui::CtxRef, frame_view: &mut FrameView) -> bool {
+    pub fn window(
+        &mut self,
+        ctx: &egui::CtxRef,
+        frame_view: &mut MaybeMutRef<'_, FrameView>,
+    ) -> bool {
         puffin::profile_function!();
         let mut open = true;
         egui::Window::new("Profiler")
@@ -364,19 +388,47 @@ impl ProfilerUi {
         if let Some(paused) = &self.paused {
             paused.selected.contains(frame_index)
         } else if let Some(latest_frame) = frame_view.latest_frame() {
-            latest_frame.frame_index == frame_index
+            latest_frame.frame_index() == frame_index
         } else {
             false
         }
     }
 
+    fn all_known_frames(&self, frame_view: &FrameView) -> Vec<Arc<FrameData>> {
+        let mut all = frame_view.all_uniq();
+        if let Some(paused) = &self.paused {
+            all.append(&mut paused.frames.all_uniq());
+        }
+        all.sort_by_key(|frame| frame.frame_index());
+        all.dedup_by_key(|frame| frame.frame_index());
+        all
+    }
+
+    fn run_pack_pass_if_needed(&mut self, frame_view: &FrameView) {
+        let last_pack_pass = self
+            .last_pack_pass
+            .get_or_insert_with(std::time::Instant::now);
+        let time_since_last_pack = last_pack_pass.elapsed();
+        if time_since_last_pack > std::time::Duration::from_secs(1) {
+            puffin::profile_scope!("pack_pass");
+            for frame in self.all_known_frames(frame_view) {
+                if !self.is_selected(frame_view, frame.frame_index()) {
+                    frame.pack();
+                }
+            }
+            self.last_pack_pass = Some(std::time::Instant::now());
+        }
+    }
+
     /// Show the profiler.
     ///
     /// Call this from within an [`egui::Window`], or use [`Self::window`] instead.
-    pub fn ui(&mut self, ui: &mut egui::Ui, frame_view: &mut FrameView) {
+    pub fn ui(&mut self, ui: &mut egui::Ui, frame_view: &mut MaybeMutRef<'_, FrameView>) {
         #![allow(clippy::collapsible_else_if)]
         puffin::profile_function!();
 
+        self.run_pack_pass_if_needed(frame_view);
+
         if !puffin::are_scopes_on() {
             ui.colored_label(ERROR_COLOR, "The puffin profiler is OFF!")
                 .on_hover_text("Turn it on with puffin::set_scopes_on(true)");
@@ -395,12 +447,27 @@ impl ProfilerUi {
                 hovered_frame = self.show_frames(ui, frame_view);
             });
 
-        let frames = if let Some(hovered_frame) = hovered_frame {
-            SelectedFrames::try_from_vec(vec![hovered_frame])
+        let frames = if let Some(frame) = hovered_frame {
+            match frame.unpacked() {
+                Ok(frame) => SelectedFrames::try_from_vec(vec![frame]),
+                Err(err) => {
+                    ui.colored_label(
+                        ERROR_COLOR,
+                        format!("Failed to load hovered frame: {}", err),
+                    );
+                    return;
+                }
+            }
         } else if let Some(paused) = &self.paused {
             Some(paused.selected.clone())
-        } else if let Some(latest_frame) = frame_view.latest_frame() {
-            SelectedFrames::try_from_vec(vec![latest_frame])
+        } else if let Some(frame) = frame_view.latest_frame() {
+            match frame.unpacked() {
+                Ok(frame) => SelectedFrames::try_from_vec(vec![frame]),
+                Err(err) => {
+                    ui.colored_label(ERROR_COLOR, format!("Failed to load latest frame: {}", err));
+                    return;
+                }
+            }
         } else {
             None
         };
@@ -412,8 +479,6 @@ impl ProfilerUi {
             return;
         };
 
-        // TODO: show age of data
-
         ui.horizontal(|ui| {
             let play_pause_button_size = Vec2::splat(24.0);
             if self.paused.is_some() {
@@ -435,10 +500,12 @@ impl ProfilerUi {
                     {
                         let latest = frame_view.latest_frame();
                         if let Some(latest) = latest {
-                            self.pause_and_select(
-                                frame_view,
-                                SelectedFrames::from_vec1(vec1::vec1![latest]),
-                            );
+                            if let Ok(latest) = latest.unpacked() {
+                                self.pause_and_select(
+                                    frame_view,
+                                    SelectedFrames::from_vec1(vec1::vec1![latest]),
+                                );
+                            }
                         }
                     }
                 });
@@ -470,7 +537,7 @@ impl ProfilerUi {
     fn show_frames(
         &mut self,
         ui: &mut egui::Ui,
-        frame_view: &mut FrameView,
+        frame_view: &mut MaybeMutRef<'_, FrameView>,
     ) -> Option<Arc<FrameData>> {
         puffin::profile_function!();
 
@@ -508,8 +575,10 @@ impl ProfilerUi {
                 ui.style_mut().wrap = Some(false);
                 ui.add_space(16.0); // make it a bit more centered
                 ui.label("Slowest:");
-                if ui.button("Clear").clicked() {
-                    frame_view.clear_slowest();
+                if let Some(frame_view) = frame_view.as_mut() {
+                    if ui.button("Clear").clicked() {
+                        frame_view.clear_slowest();
+                    }
                 }
             });
 
@@ -536,6 +605,26 @@ impl ProfilerUi {
             });
         });
 
+        {
+            let uniq = frames.all_uniq();
+            let mut bytes = 0;
+            let mut unpacked = 0;
+            for frame in &uniq {
+                bytes += frame.bytes_of_ram_used();
+                unpacked += frame.has_unpacked() as usize;
+            }
+            ui.label(format!(
+                "{} frames ({} unpacked) using approximately {:.1} MB.",
+                uniq.len(),
+                unpacked,
+                bytes as f64 * 1e-6
+            ));
+
+            if let Some(frame_view) = frame_view.as_mut() {
+                max_frames_ui(ui, frame_view);
+            }
+        }
+
         hovered_frame
     }
 
@@ -555,7 +644,7 @@ impl ProfilerUi {
             frames.len() as f32 * frame_width_including_spacing
         } else {
             // leave gaps in the view for the missing frames
-            let num_frames = frames[frames.len() - 1].frame_index + 1 - frames[0].frame_index;
+            let num_frames = frames[frames.len() - 1].frame_index() + 1 - frames[0].frame_index();
             num_frames as f32 * frame_width_including_spacing
         };
 
@@ -579,9 +668,9 @@ impl ProfilerUi {
             let x = if tight {
                 rect.right() - (frames.len() as f32 - i as f32) * frame_width_including_spacing
             } else {
-                let latest_frame_index = frames[frames.len() - 1].frame_index;
+                let latest_frame_index = frames[frames.len() - 1].frame_index();
                 rect.right()
-                    - (latest_frame_index + 1 - frame.frame_index) as f32
+                    - (latest_frame_index + 1 - frame.frame_index()) as f32
                         * frame_width_including_spacing
             };
 
@@ -594,7 +683,7 @@ impl ProfilerUi {
                 let duration = frame.duration_ns();
                 slowest_visible_frame = duration.max(slowest_visible_frame);
 
-                let is_selected = self.is_selected(frame_view, frame.frame_index);
+                let is_selected = self.is_selected(frame_view, frame.frame_index());
 
                 let is_hovered = if let Some(mouse_pos) = response.hover_pos() {
                     response.hovered()
@@ -627,7 +716,9 @@ impl ProfilerUi {
                         let max_x = start.x.max(curr.x);
                         let intersects = min_x <= frame_rect.right() && frame_rect.left() <= max_x;
                         if intersects {
-                            new_selection.push(frame.clone());
+                            if let Ok(frame) = frame.unpacked() {
+                                new_selection.push(frame);
+                            }
                         }
                     }
                 }
@@ -665,55 +756,38 @@ impl ProfilerUi {
 fn frames_info_ui(ui: &mut egui::Ui, selection: &SelectedFrames) {
     let mut sum_ns = 0;
     let mut sum_scopes = 0;
-    let mut sum_bytes = 0;
-
-    let mut knows_compressed_size = true;
-    let mut sum_bytes_compressed = 0;
 
     for frame in &selection.frames {
-        let (min_ns, max_ns) = frame.range_ns;
+        let (min_ns, max_ns) = frame.range_ns();
         sum_ns += max_ns - min_ns;
 
-        sum_scopes += frame.num_scopes;
-        sum_bytes += frame.num_bytes;
-
-        if let Some(compressed_size) = frame.compressed_size {
-            sum_bytes_compressed += compressed_size;
-        } else {
-            knows_compressed_size = false;
-        }
+        sum_scopes += frame.meta.num_scopes;
     }
 
     let frame_indices = if selection.frames.len() == 1 {
-        format!("frame #{}", selection.frames[0].frame_index)
+        format!("frame #{}", selection.frames[0].frame_index())
     } else if selection.frames.len() as u64
-        == selection.frames.last().frame_index - selection.frames.first().frame_index + 1
+        == selection.frames.last().frame_index() - selection.frames.first().frame_index() + 1
     {
         format!(
             "{} frames (#{} - #{})",
             selection.frames.len(),
-            selection.frames.first().frame_index,
-            selection.frames.last().frame_index
+            selection.frames.first().frame_index(),
+            selection.frames.last().frame_index()
         )
     } else {
         format!("{} frames", selection.frames.len())
     };
 
     let mut info = format!(
-        "Showing {}, {:.1} ms, {} threads, {} scopes, {:.1} kB",
+        "Showing {}, {:.1} ms, {} threads, {} scopes.",
         frame_indices,
         sum_ns as f64 * 1e-6,
         selection.threads.len(),
         sum_scopes,
-        sum_bytes as f64 * 1e-3
     );
-
-    if knows_compressed_size {
-        info += &format!(" ({:.1} kB compressed)", sum_bytes_compressed as f64 * 1e-3);
-    }
-
     if let Some(time) = format_time(selection.raw_range_ns.0) {
-        info += &format!(". Recorded {}", time);
+        info += &format!(" Recorded {}.", time);
     }
 
     ui.label(info);
@@ -729,3 +803,35 @@ fn format_time(nanos: NanoSecond) -> Option<String> {
         None // `nanos` is likely not counting from epoch.
     }
 }
+
+fn max_frames_ui(ui: &mut egui::Ui, frame_view: &mut FrameView) {
+    let uniq = frame_view.all_uniq();
+
+    let mut bytes = 0;
+    for frame in &uniq {
+        bytes += frame.bytes_of_ram_used();
+    }
+
+    let frames_per_second = if let (Some(first), Some(last)) = (uniq.first(), uniq.last()) {
+        let nanos = last.range_ns().1 - first.range_ns().0;
+        let seconds = nanos as f64 * 1e-9;
+        let frames = last.frame_index() - first.frame_index() + 1;
+        frames as f64 / seconds
+    } else {
+        60.0
+    };
+
+    ui.horizontal(|ui| {
+        ui.label("Max recent frames to store:");
+
+        let mut memory_length = frame_view.max_recent();
+        ui.add(egui::Slider::new(&mut memory_length, 10..=100_000).logarithmic(true));
+        frame_view.set_max_recent(memory_length);
+
+        ui.label(format!(
+            "(≈ {:.1} minutes, ≈ {:.0} MB)",
+            memory_length as f64 / 60.0 / frames_per_second,
+            memory_length as f64 * bytes as f64 / uniq.len() as f64 * 1e-6,
+        ));
+    });
+}
